/*
 *
 * written by Simon Barrick, 2007
 * simon@fourthwall.ndo.co.uk
 *
 * Warp SOP
 *
 * this sop is one of two sops to capture and deform an input mesh
 * this part captures points using a mean value coordinate system, calculated
 * using spherical barycentric coordinates.
 * It is based on a paper by Torsten Langer, Alexander Belyaev and Hans-Peter Seidel
 *
 */
#define halfPI 1.570796326794
#define MAXFLT -65535

//#define BETA
#define OBJECTMERGESOP_MENU ((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE | \
                                              PRM_CHOICELIST_REPLACE))
#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <UT/UT_Matrix3.h>
#include <GU/GU_Detail.h>
#include <GU/GU_PrimPoly.h>
#include <PRM/PRM_Include.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <SOP/SOP_Guide.h>
#include <UT/UT_PtrArray.h>
#include <UT/UT_ErrorManager.h>
#include <OP/OP_Director.h>
#include <UT/UT_Vector3.h>
#include <UT/UT_Interrupt.h>
#include <GEO/GEO_Detail.h>
#include <GEO/GEO_Point.h>
#include <GEO/GEO_Vertex.h>
#include <GB/GB_Attribute.h>
#include <GB/GB_AttributeTable.h>
#include <GB/GB_ExtraMacros.h>
#include <SYS/fpreal16.h>
#include <math.h>
#include <iostream>
//#include <vector>

#include "./SOP_SBC_capt.H"


// This is were the code switches between the standard
// compile and the beta compile.
#ifdef BETA
static const char* const SOP_INTERNAL_NAME = "SBC_capt_beta";
static const char* const SOP_UI_NAME = "SBC capt Beta";
#else
static const char* const SOP_INTERNAL_NAME = "SBC_capt";
static const char* const SOP_UI_NAME = "SBC capt";
#endif


void
newSopOperator( OP_OperatorTable *table )
{
  table->addOperator
  (
    new OP_Operator
    (
      SOP_INTERNAL_NAME,             // Internal name
      SOP_UI_NAME,                   // UI name
      SOP_SBC_capt::myConstructor,  //
      SOP_SBC_capt::myTemplateList, //
      2,                             // Min Inputs
      2,                             // Max Inputs
      0                              // Local Variables
      //OP_FLAG_*                    // Additional Flag
    )

// Examples of additional flags include, OP_FLAG_GENERATOR, OP_FLAG_UNORDERED.
// See OP_Operator.h for more details.

  );
}

static PRM_Name names[] =
{
  PRM_Name( "pre_convexRest", "Pre Convex Rest" ),

  PRM_Name( "accurate_weights", "Accurate weights on large objects" ),

  PRM_Name( "full_checks", "Full checks for points coincident with rest mesh" ),

  PRM_Name( "usethres", "Use threshold"),

  PRM_Name( "thres", "Threshold"),

  PRM_Name( 0 )
};

static PRM_Default  smallDefault(0.1);   // Default to 0.1

// How to make choices for a menu.

PRM_Template
SOP_SBC_capt::myTemplateList[] =
{
  PRM_Template(PRM_TOGGLE,    1, &names[0]),

  //PRM_Template(PRM_TOGGLE,    1, &names[1]),

  //PRM_Template(PRM_TOGGLE,    1, &names[2]),

  PRM_Template(PRM_TOGGLE,    1, &names[3]),

  PRM_Template(PRM_FLT,       1, &names[4], &smallDefault, 0, &PRMunitRange),

  //PRM_Template(PRM_CALLBACK, 1, &names[1], 0, 0, 0, SOP_SBC_capt::launchManualCallback ),

  PRM_Template(),
};

const char *
SOP_SBC_capt::inputLabel( unsigned int input) const
{
 switch (input) {
   case 0:
     return "Geometry To Deform";
     break;
   case 1:
     return "Rest Mesh";
     break;
   default:
     return "WTF?";
     break;
 }
}

//overriding constructor
OP_Node *
SOP_SBC_capt::myConstructor( OP_Network *net, const char *name, OP_Operator *op )
{
  return new SOP_SBC_capt( net, name, op );
}

SOP_SBC_capt::SOP_SBC_capt( OP_Network *net, const char *name, OP_Operator *op )
  : SOP_Node( net, name, op )
{

    mySopFlags.setNeedGuide1( 1 );

}

//destructor
SOP_SBC_capt::~SOP_SBC_capt() {}

unsigned
SOP_SBC_capt::disableParms()
{
    unsigned changed = 0;

    changed  = enableParm("thres", USETHRES());

    return changed;
}

int
SOP_SBC_capt::cyclicMod(int v, int n)
{
  return (v==-1?(n-1):(v==n?0:v));
}

float
SOP_SBC_capt::calcAlpha(const GEO_Primitive *prim, int i, UT_Vector3 vF, UT_Vector3 pos, float theta)
{

  vF.normalize();

  int vertexCount = prim->getVertexCount();

  GEO_Point *ppt = prim->getVertex(i).getPt();

  UT_Vector3 v1 = ppt->getPos()-pos;
  v1.normalize();

  UT_Vector3 vcross1 = v1;
  vcross1.cross(vF);
  vcross1.normalize();

  const GEO_Vertex *vtx = &prim->getVertex(cyclicMod(i+1,vertexCount));
  ppt = vtx->getPt();

  UT_Vector3 v2 = ppt->getPos()-pos;
  v2.normalize();

  UT_Vector3 vcross2;

  float alpha = acos(vcross1.dot(vcross2));
  if (theta > halfPI)
  {
    vcross2 = -v2;
    vcross2.cross(vF);
    vcross2.normalize();
    alpha = acos(vcross1.dot(vcross2));
  }
  else
  {
    vcross2 = v2;
    vcross2.cross(vF);
    vcross2.normalize();
    alpha = acos(vcross1.dot(vcross2));
  }

  return alpha;

}

float
SOP_SBC_capt::calcTheta(UT_Vector3 pos, const GEO_Primitive *prim, int i, UT_Vector3 vF)
{
  //calc angle between v_i and vF

  int vertexCount = prim->getVertexCount();

  GEO_Point *ppt = prim->getVertex(i).getPt();
  UT_Vector3 u0 = (ppt->getPos()-pos);
  u0.normalize();

  vF.normalize();

  return (acos(u0.dot(vF)));

}

UT_Vector3
SOP_SBC_capt::calcFaceNormal(UT_Vector3 pos, const GEO_Primitive *prim)
{
//vF = sum 0.5 angle between u and nextu multiplied by n
//n = cross(u,nextu) / length(cross(u,nextu)

  UT_Vector3 vF(0,0,0);

  int vertexCount = prim->getVertexCount();

  UT_Vector3    u0,u1;

  GEO_Point *ppt = prim->getVertex(0).getPt();
  u0 = (ppt->getPos()-pos);
  u0.normalize();

  for(int i=0; i<vertexCount; i++)
  {
    const GEO_Vertex *vtx = &prim->getVertex(cyclicMod(i+1,vertexCount));

    ppt = vtx->getPt();
    u1 = (ppt->getPos()-pos);
    u1.normalize();

    UT_Vector3 ucross = u0;
    ucross.cross(u1);

    UT_Vector3 n = ucross;
    n.normalize();
    float theta = acos(u0.dot(u1));

    vF += 0.5 * theta * n;

    u0 = u1;

  }

  return vF;

}

float
SOP_SBC_capt::calcWeights(UT_Vector3 pos, const GEO_Primitive *prim, UT_FloatArray &numerators)
{
  //calculate face normal vF
  UT_Vector3 vF = calcFaceNormal(pos,prim);
  int vertexCount = prim->getVertexCount();

  float denominator = 0;

  float alpha1 = calcAlpha(prim, vertexCount-1, vF, pos, 0);

  for(int i=0; i<vertexCount; i++)
  {
    GEO_Point *ppt = prim->getVertex(i).getPt();
    //f1
    float f1 = (vF.length()) / ((ppt->getPos()-pos).length());
    float theta = calcTheta(pos, prim, i, vF);
    float alpha2 = calcAlpha(prim, i, vF, pos, 0);
    float f2 = tan(alpha1/2) + tan(alpha2/2);

    denominator += (1/tan(theta))*f2;
    numerators(i+1) = f1*(f2/sin(theta));

    alpha1 = alpha2;
  }

  return denominator;
}

//cook method

OP_ERROR
SOP_SBC_capt::cookMySop( OP_Context &context )
{
  int                   i,j,attrib,nptattrib;
  float                 now = context.myTime;
  double                total_weight;
  float                 w;
  GEO_Point             *ppt,*pt;
  const GEO_Primitive   *prim;
  const GEO_Vertex      *vtx;
  UT_Interrupt          *boss;
  GU_Detail             *rest;
  GU_Detail             restMesh;
  float                 *captatt;
  static int            maxrp;
  UT_BoundingBox        bbox;

  //check for user interrupts
  boss = UTgetInterrupt();

  boss->opStart("Capturing points");

  // Before we do anything, we must lock our inputs.  Before returning,
  // we have to make sure that the inputs get unlocked.
  if( lockInputs( context ) >= UT_ERROR_ABORT ) return error();

  duplicateSource( 1, context, &restMesh );
  rest = &restMesh;

  if (PRECONVEX()) {
    rest->convex(100);
  }

  //copy input geomtry as we aren't creating any
  duplicateChangedSource( 0, context );

//sparse stuff
  int npts = gdp->points().entries();
  //create an array of UT_FloatArrays pointed at by dataPairs
  UT_IntArray       *dataArray=new UT_IntArray[npts];
  UT_FloatArray     *weightArray=new UT_FloatArray[npts];
//

  maxrp = rest->points().entries();

  int detoff = gdp->findAttrib("sbcCaptn", GB_ATTRIB_INT);
  if (detoff > -1)
  {
    int *detailatt = (int*)gdp->attribs().getAttribData(detoff);

    gdp->destroyAttrib("sbcCaptn",sizeof(int),GB_ATTRIB_INT);

    gdp->destroyPointAttrib("sbcCapt",detailatt[0]*sizeof(float)+sizeof(float),GB_ATTRIB_FLOAT);
    gdp->destroyPointAttrib("sbcCapt",sizeof(float),GB_ATTRIB_FLOAT);
    gdp->destroyPtIndexPairAttribute("sbcCaptp");
  }

  // Capture - create and initialise attribrites
  if (USETHRES())
  {
    attrib = gdp->addPointAttrib("sbcCapt",sizeof(float),GB_ATTRIB_FLOAT,0);
  }
  else
  {
    attrib = gdp->addPointAttrib("sbcCapt",maxrp*sizeof(float)+sizeof(float),GB_ATTRIB_FLOAT,0);

    nptattrib = gdp->addAttrib("sbcCaptn",sizeof(int),GB_ATTRIB_INT,&maxrp);
  }

  //if (ACC())

  rest->getBBox(&bbox);
  UT_Vector3 pos;
  FOR_ALL_GPOINTS(rest,ppt)
  {
    pos = ppt->getPos();
    ppt->getPos() = pos/bbox.sizeMax();
  }

  bool goodCapture = true;

  //Iterate all prims in the rest geometry for all points in the input
  FOR_ALL_GPOINTS(gdp,ppt)
  {
    //get current point coord
    pos = ppt->getPos();
    //if (ACC())
    pos /= bbox.sizeMax();
    captatt = (float*)ppt->getAttribData(attrib);

    //initial
    total_weight = 0;

    //check for user interrupts
    if (boss->opInterrupt())
      break;

    UT_FloatArray numerators(10);

    FOR_ALL_PRIMITIVES(rest,prim)
    {
      int vertexCount = prim->getVertexCount();
      //calculate w
      //numerators.entries(vertexCount);
      float totalDenominator = calcWeights(pos,prim,numerators);

      for(i=0; i<vertexCount;i++)
      {
        if (absolute(totalDenominator) > 0.0001)
          w = numerators(i+1)/totalDenominator;
        else
        {
          if (goodCapture)
            addWarning(SOP_MESSAGE , "Capture geometry very close to cage, may result in bad capture");
          goodCapture = false;
          w = 0;
        }

        if (USETHRES())
          w=absolute(w)<THRES()?0:w;

        w *= -1;

        total_weight += w;

        vtx = &prim->getVertex(i);
        pt=vtx->getPt();
        int ptnum=pt->getNum();

        if (USETHRES() && w)
        {
          int offset = dataArray[ppt->getNum()].find(ptnum);
          if (offset<0)
          {
            dataArray[ppt->getNum()].append(ptnum);
            weightArray[ppt->getNum()].append(w);
          }
          else
          {
            weightArray[ppt->getNum()](offset)+=w;
          }
        }
        else
        {
          captatt[ptnum] += w;
        }
      }
    }

    //store total_weight
    if (USETHRES())
      captatt[0] = (float)total_weight;
    else
      captatt[rest->points().entries()] = (float)total_weight;
  }

  int maxCount=0;
  if (USETHRES())
  {
    for(i=0;i<npts;i++)
    {
      if ( dataArray[i].entries() > maxCount )
        maxCount = dataArray[i].entries();
    }

    nptattrib = gdp->addAttrib("sbcCaptn",sizeof(int),GB_ATTRIB_INT,&maxCount);

    attrib = gdp->addPtIndexPairAttribute("sbcCaptp",maxCount*sizeof(float)*2);
    //initialise to zero instead of -1
    FOR_ALL_GPOINTS(gdp,ppt)
    {
      captatt = (float*)ppt->getAttribData(attrib);
      for(j=0;j<maxCount*2;j++)
      {
        captatt[j] = 0;
      }
    }

    i = 0;
    FOR_ALL_GPOINTS(gdp,ppt)
    {
      captatt = (float*)ppt->getAttribData(attrib);
      for(j=0;j<dataArray[i].entries();j++)
      {
        captatt[j*2] = dataArray[i][j];
        captatt[j*2+1] = weightArray[i][j];
      }
      i++;
    }
  }


  delete[] dataArray;
  delete[] weightArray;

  unlockInputs();

  boss->opEnd();

  return error();

}


OP_ERROR
SOP_SBC_capt::cookMyGuide1( OP_Context &context )
{
  float now = context.myTime;

  if ( lockInputs( context ) >= UT_ERROR_ABORT ) return error();

  GU_Detail *in_gdp = (GU_Detail *)inputGeo(1, context);
  if (in_gdp)
    myGuide1->copy(*in_gdp);

  unlockInputs();
  return error();
}

