/*
 *
 * written by Simon Barrick, 2007
 * simon@fourthwall.ndo.co.uk
 *
 * SBCdeform SOP
 *
 * this sop is one of two sops to capture and deform an input mesh
 * this part deforms points that have been captured using a mean value coordinate system.
 *
 */

//#define BETA
#define OBJECTMERGESOP_MENU ((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE | \
                                              PRM_CHOICELIST_REPLACE))

#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <UT/UT_Matrix3.h>
#include <GU/GU_Detail.h>
#include <GU/GU_PrimPoly.h>
#include <PRM/PRM_Include.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <SOP/SOP_Guide.h>
#include <UT/UT_PtrArray.h>
#include <UT/UT_ErrorManager.h>
#include <OP/OP_Director.h>
#include <UT/UT_Vector3.h>
#include <UT/UT_Interrupt.h>
#include <GEO/GEO_Detail.h>
#include <GEO/GEO_Point.h>
#include <GEO/GEO_Vertex.h>
#include <GB/GB_Attribute.h>
#include <GB/GB_AttributeTable.h>
#include <GB/GB_AttributeHandle.h>

#include <math.h>
#include <iostream>
//#include <vector>

#include "./SOP_SBC_deform.H"

// This is were the code switches between the standard
// compile and the beta compile.
#ifdef BETA
static const char* const SOP_INTERNAL_NAME = "SBC_deform_beta";
static const char* const SOP_UI_NAME = "SBC deform Beta";
#else
static const char* const SOP_INTERNAL_NAME = "SBC_deform";
static const char* const SOP_UI_NAME = "SBC deform";
#endif



void
newSopOperator( OP_OperatorTable *table )
{
  table->addOperator
  (
    new OP_Operator
    (
      SOP_INTERNAL_NAME,             // Internal name
      SOP_UI_NAME,                   // UI name
      SOP_SBC_deform::myConstructor,  //
      SOP_SBC_deform::myTemplateList, //
      2,                             // Min Inputs
      2,                             // Max Inputs
      0                              // Local Variables
      //OP_FLAG_*                    // Additional Flag
    )

// Examples of additional flags include, OP_FLAG_GENERATOR, OP_FLAG_UNORDERED.
// See OP_Operator.h for more details.

  );
}

static PRM_Name names[] =
{
  PRM_Name( "delatt", "Delete capture attributes" ),
  PRM_Name( 0 )
};

// How to make choices for a menu.

PRM_Template
SOP_SBC_deform::myTemplateList[] =
{
  PRM_Template(PRM_TOGGLE,    1, &names[0]),
  PRM_Template(),
};

const char *
SOP_SBC_deform::inputLabel( unsigned int input) const
{
 switch (input) {
   case 0:
     return "Geometry To Deform";
     break;
   case 1:
     return "Deformed Mesh";
     break;
   default:
     return "WTF?";
     break;
 }
}

//overriding constructor
OP_Node *
SOP_SBC_deform::myConstructor( OP_Network *net, const char *name, OP_Operator *op )
{
  return new SOP_SBC_deform( net, name, op );
}

SOP_SBC_deform::SOP_SBC_deform( OP_Network *net, const char *name, OP_Operator *op )
  : SOP_Node( net, name, op )
{

    mySopFlags.setNeedGuide1( 1 );

}

//destructor
SOP_SBC_deform::~SOP_SBC_deform() {}

//don't even disable or enable anything
unsigned
SOP_SBC_deform::disableParms()
{
    return 0;
}


//cook method

OP_ERROR
SOP_SBC_deform::cookMySop( OP_Context &context )
{

  int                   i,j,attrib,attribtw,detoff;
  float                 now = context.myTime;
  float                 total_weight,weight;
  float                 w;
  UT_Vector3            total_function;
  GEO_Point             *ppt;
  const GEO_Point       *pt;
  UT_Vector3            pos;
  const GU_Detail       *deform;
  float                 *captatt,*captatttw;
  static int            maxrp;
  int                   *detailatt;
  bool                  treatAsPairs=false;


  // Before we do anything, we must lock our inputs.  Before returning,
  // we have to make sure that the inputs get unlocked.
  if( lockInputs( context ) >= UT_ERROR_ABORT ) return error();
  if( (deform =   inputGeo(1,context)) == NULL ) return error();

  //copy input geomtry as we aren't creating any
  duplicateSource( 0, context );

  detoff = gdp->findAttrib("sbcCaptn", GB_ATTRIB_INT);
  detailatt = (int*)gdp->attribs().getAttribData(detoff);

  if (detoff<0) {
     addError(SOP_ERR_INVALID_SRC , "input 1 is not captured");
     unlockInputs();
     return error();
  }
  else {
    if ( *detailatt != deform->points().entries())
      treatAsPairs = true;
  }

  //deform only
  if (treatAsPairs)
  {
    attrib = gdp->findPtIndexPairAttribute("sbcCaptp");

    if( attrib >= 0 )
    {
      for ( i = 0; i < gdp->points().entries(); i++ )
      {
        ppt = gdp->points()( i );
        captatt = (float*)ppt->getAttribData(attrib);
        total_function = (0,0,0);
        for ( j = 0; j < *detailatt; j++ )
        {
          //get pointer to current point
          pt = deform->points()(captatt[j*2]);
          weight = captatt[j*2+1];
          //if (!weight) break;
          //get current point coord
          pos = pt->getPos();
          total_function += pos*weight;
        }

        attribtw = gdp->findPointAttrib("sbcCapt",GB_ATTRIB_FLOAT);
        captatttw = (float*)ppt->getAttribData(attribtw);
        total_weight = captatttw[0];
        //normalise final position
        pos = total_function/total_weight;
        //set new coord
        ppt->getPos() = pos;
      }
    }
    else {
      addError(SOP_ERR_INVALID_SRC , "input 1 is not captured");
      unlockInputs();
      return error();
    }
  }
  else
  {
    attrib = gdp->findPointAttrib("sbcCapt",GB_ATTRIB_FLOAT);

    if( attrib >= 0 )
    {
      for ( i = 0; i < gdp->points().entries(); i++ )
      {
        ppt = gdp->points()( i );
        captatt = (float*)ppt->getAttribData(attrib);
        total_function = (0,0,0);
        for ( j = 0; j < deform->points().entries(); j++ )
        {
          //get pointer to current point
          pt = deform->points()( j );
          //get current point coord
          pos = pt->getPos();
          total_function += pos*captatt[j];
        }

        total_weight = captatt[deform->points().entries()];
        //normalise final position
        pos = total_function/total_weight;
        //set new coord
        ppt->getPos() = pos;
      }
    }
    else {
      addError(SOP_ERR_INVALID_SRC , "input 1 is not captured");
      unlockInputs();
      return error();
    }
  }

  if (DELATT()) {
    maxrp = deform->points().entries();
    gdp->destroyPointAttrib("sbcCapt",maxrp*sizeof(float)+sizeof(float),GB_ATTRIB_FLOAT);
    gdp->destroyPointAttrib("sbcCapt",sizeof(float),GB_ATTRIB_FLOAT);
    gdp->destroyAttrib("sbcCaptn",sizeof(int),GB_ATTRIB_INT);
    gdp->destroyPtIndexPairAttribute("sbcCaptp");
  }

  unlockInputs();


  return error();

}


OP_ERROR
SOP_SBC_deform::cookMyGuide1( OP_Context &context )
{
  float now = context.myTime;

  if ( lockInputs( context ) >= UT_ERROR_ABORT ) return error();

  GU_Detail *in_gdp = (GU_Detail *)inputGeo(1, context);
  if (in_gdp)
    myGuide1->copy(*in_gdp);

  unlockInputs();
  return error();
}


