/*
 *
 * written by Simon Barrick, 2007
 * simon@fourthwall.ndo.co.uk
 *
 * SBCdeform SOP
 *
 * this sop is one of two sops to capture and cage an input mesh
 * this part deforms points that have been captured using a mean value coordinate system.
 *
 */

//#define BETA
#define OBJECTMERGESOP_MENU ((PRM_ChoiceListType)(PRM_CHOICELIST_EXCLUSIVE | \
                                              PRM_CHOICELIST_REPLACE))

#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <UT/UT_Matrix3.h>
#include <GU/GU_Detail.h>
#include <GU/GU_PrimPoly.h>
#include <PRM/PRM_Include.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <SOP/SOP_Guide.h>
#include <UT/UT_PtrArray.h>
#include <UT/UT_ErrorManager.h>
#include <OP/OP_Director.h>
#include <UT/UT_Vector3.h>
#include <UT/UT_Interrupt.h>
#include <GEO/GEO_Detail.h>
#include <GEO/GEO_Point.h>
#include <GEO/GEO_Vertex.h>
#include <GB/GB_Attribute.h>
#include <GB/GB_AttributeTable.h>
#include <GB/GB_AttributeHandle.h>
#include <UT/UT_WorkArgs.h>

#include <math.h>
#include <iostream>
//#include <vector>

#include "./SOP_SBC_transfer.H"

// This is were the code switches between the standard
// compile and the beta compile.
#ifdef BETA
static const char* const SOP_INTERNAL_NAME = "SBC_transfer_beta";
static const char* const SOP_UI_NAME = "SBC transfer Beta";
#else
static const char* const SOP_INTERNAL_NAME = "SBC_transfer";
static const char* const SOP_UI_NAME = "SBC transfer";
#endif



void
newSopOperator( OP_OperatorTable *table )
{
  table->addOperator
  (
    new OP_Operator
    (
      SOP_INTERNAL_NAME,             // Internal name
      SOP_UI_NAME,                   // UI name
      SOP_SBC_transfer::myConstructor,  //
      SOP_SBC_transfer::myTemplateList, //
      2,                             // Min Inputs
      2,                             // Max Inputs
      0                              // Local Variables
      //OP_FLAG_*                    // Additional Flag
    )

// Examples of additional flags include, OP_FLAG_GENERATOR, OP_FLAG_UNORDERED.
// See OP_Operator.h for more details.

  );
}

static PRM_Name names[] =
{
  PRM_Name( "delatt", "Delete capture attributes" ),
  PRM_Name("attribs", "Point attributes to transfer"),
  PRM_Name("norm", "Normalise vector attributes"),
  PRM_Name( 0 )
};

// How to make choices for a menu.

PRM_Template
SOP_SBC_transfer::myTemplateList[] =
{
  PRM_Template(PRM_TOGGLE,    1, &names[0]),

  PRM_Template(PRM_STRING,    1, &names[1]),

  PRM_Template(PRM_TOGGLE,    1, &names[2]),

  PRM_Template(),
};

const char *
SOP_SBC_transfer::inputLabel( unsigned int input) const
{
 switch (input) {
   case 0:
     return "Geometry to transfer attributes to";
     break;
   case 1:
     return "Geometry to transfer attributes from";
     break;
   default:
     return "WTF?";
     break;
 }
}

//overriding constructor
OP_Node *
SOP_SBC_transfer::myConstructor( OP_Network *net, const char *name, OP_Operator *op )
{
  return new SOP_SBC_transfer( net, name, op );
}

SOP_SBC_transfer::SOP_SBC_transfer( OP_Network *net, const char *name, OP_Operator *op )
  : SOP_Node( net, name, op )
{

    mySopFlags.setNeedGuide1( 1 );

}

//destructor
SOP_SBC_transfer::~SOP_SBC_transfer() {}

//don't even disable or enable anything
unsigned
SOP_SBC_transfer::disableParms()
{
    return 0;
}


//cook method

OP_ERROR
SOP_SBC_transfer::cookMySop( OP_Context &context )
{

  int                   i,j,captAttribOff,captAttribTWoff,detoff,currentAttribOff,newAttribOff,floatind;
  float                 now = context.myTime;
  float                 totalWeight,weight;
  float                 w;
  GEO_Point             *ppt;
  const GEO_Point       *pt;
  UT_Vector3            currentVectorAtt, vectorAtt;
  const GU_Detail       *cage;
  float                 *captAtt,*captAttTW, *currentFloatAtt, *newFloatAttrib;
  static int            maxrp;
  int                   *detailatt;
  bool                  treatAsPairs=false;
  GB_Attribute          *attrib;
  UT_FloatArray         totalFloatFunction(1);

  // Before we do anything, we must lock our inputs.  Before returning,
  // we have to make sure that the inputs get unlocked.
  if( lockInputs( context ) >= UT_ERROR_ABORT ) return error();
  if( (cage =   inputGeo(1,context)) == NULL ) return error();

  //copy input geomtry as we aren't creating any
  duplicateSource( 0, context );

  detoff = gdp->findAttrib("sbcCaptn", GB_ATTRIB_INT);
  detailatt = (int*)gdp->attribs().getAttribData(detoff);

  if (detoff<0) {
     addError(SOP_ERR_INVALID_SRC , "input 1 is not captured");
     unlockInputs();
     return error();
  }
  else {
    if ( *detailatt != cage->points().entries())
      treatAsPairs = true;
  }

  UT_String AttString;
  getATTS(AttString);
  AttString.trimSpace(1);
  UT_WorkArgs Args;
  AttString.tokenize (Args);

  bool normalise = NORM();

  if (treatAsPairs)
    captAttribOff = gdp->findPtIndexPairAttribute("sbcCaptp");
  else
    captAttribOff = gdp->findPtIndexPairAttribute("sbcCapt");

  if( captAttribOff >= 0 )
  {
    //Loop though all attrs
    for(int  argNum = 0;argNum<Args.getArgc();argNum++)
    {
      int type = -1;
      attrib = cage->pointAttribs().find((const char *)Args[argNum], GB_ATTRIB_FLOAT);
      currentAttribOff = cage->findPointAttrib((const char *)Args[argNum], GB_ATTRIB_FLOAT);
      if(attrib)
      {
        type = 0;
        newAttribOff = gdp->findPointAttrib((const char *)Args[argNum], GB_ATTRIB_FLOAT);
        if ( newAttribOff < 0 )
          newAttribOff = gdp->addPointAttrib((const char *)Args[argNum],attrib->getSize(),GB_ATTRIB_FLOAT,0);
      }
      else
      {
        attrib = cage->pointAttribs().find((const char *)Args[argNum], GB_ATTRIB_VECTOR);
        currentAttribOff = cage->findPointAttrib((const char *)Args[argNum], GB_ATTRIB_VECTOR);
        if(attrib)
        {
          type = 1;
          newAttribOff = gdp->findPointAttrib((const char *)Args[argNum], GB_ATTRIB_VECTOR);
          if ( newAttribOff < 0 )
            newAttribOff = gdp->addPointAttrib((const char *)Args[argNum],attrib->getSize(),GB_ATTRIB_VECTOR,0);
        }
      }

      if(type == -1) continue;

      for ( i = 0; i < gdp->points().entries(); i++ )
      {
        ppt = gdp->points()( i );
        captAtt = (float*)ppt->getAttribData(captAttribOff);
        totalFloatFunction.entries(attrib->getSize()/sizeof(float));
        for (floatind = 0; floatind < attrib->getSize()/sizeof(float); floatind++)
          totalFloatFunction[floatind+1] = 0;

        UT_Vector3      totalVectorFunction(0,0,0);

        int loops;
        if (treatAsPairs)
          loops = *detailatt;
        else
          loops = cage->points().entries();

        for ( j = 0; j < loops; j++ )
        {
          //get pointer to current point
          if (treatAsPairs)
          {
            pt = cage->points()(captAtt[j*2]);
            weight = captAtt[j*2+1];
          }
          else
          {
            pt = cage->points()( j );
            weight = captAtt[j];
          }

          //get current point attribute
          if (type == 0)
          {
            currentFloatAtt = (float*)pt->getAttribData(currentAttribOff);
            for (floatind = 0; floatind < attrib->getSize()/sizeof(float); floatind++)
              totalFloatFunction[floatind+1] += currentFloatAtt[floatind]*weight;
          }
          else if (type == 1)
          {
            //probably should deal with vector4 att too
            currentFloatAtt = (float*)pt->getAttribData(currentAttribOff);
            currentVectorAtt[0] = currentFloatAtt[0];
            currentVectorAtt[1] = currentFloatAtt[1];
            currentVectorAtt[2] = currentFloatAtt[2];
            totalVectorFunction += currentVectorAtt*weight;
          }
        }

        if (treatAsPairs)
        {
          captAttribTWoff = gdp->findPointAttrib("sbcCapt",GB_ATTRIB_FLOAT);
          captAttTW = (float*)ppt->getAttribData(captAttribTWoff);
          totalWeight = captAttTW[0];
        }
        else
          totalWeight = captAtt[cage->points().entries()];

        newFloatAttrib = (float*)ppt->getAttribData(newAttribOff);
        //normalise final att
        if (type == 0)
        {
          for (floatind = 0; floatind < attrib->getSize()/sizeof(float); floatind++)
          {
            if (absolute(totalWeight) > 0.0001)
              newFloatAttrib[floatind] = totalFloatFunction[floatind+1]/totalWeight;
            else
              newFloatAttrib[floatind] = 0;
          }
        }
        else if (type == 1)
        {
          vectorAtt = totalVectorFunction/totalWeight;
          if (normalise)
            vectorAtt.normalize();

          newFloatAttrib[0] = vectorAtt[0];
          newFloatAttrib[1] = vectorAtt[1];
          newFloatAttrib[2] = vectorAtt[2];
        }
      }
    }
  }
  else {
    addError(SOP_ERR_INVALID_SRC , "input 1 is not captured");
    unlockInputs();
    return error();
  }

  if (DELATT()) {
    maxrp = cage->points().entries();
    gdp->destroyPointAttrib("sbcCapt",maxrp*sizeof(float)+sizeof(float),GB_ATTRIB_FLOAT);
    gdp->destroyPointAttrib("sbcCapt",sizeof(float),GB_ATTRIB_FLOAT);
    gdp->destroyAttrib("sbcCaptn",sizeof(int),GB_ATTRIB_INT);
    gdp->destroyPtIndexPairAttribute("sbcCaptp");
  }

  unlockInputs();


  return error();

}


OP_ERROR
SOP_SBC_transfer::cookMyGuide1( OP_Context &context )
{
  float now = context.myTime;

  if ( lockInputs( context ) >= UT_ERROR_ABORT ) return error();

  GU_Detail *in_gdp = (GU_Detail *)inputGeo(1, context);
  if (in_gdp)
    myGuide1->copy(*in_gdp);

  unlockInputs();
  return error();
}


